@@ -15,20 +15,22 @@
 #include "config.h"
 #endif
 
-/*#define DEBUG*/
-
 #include <string.h>
 #include <FLAC/all.h>
 #include <stdlib.h>
 #include <strings.h>
+
+/*#define DEBUG*/
+
+#include "common.h"
 #include "audio.h"
 #include "decoder.h"
 #include "server.h"
-#include "common.h"
 #include "log.h"
 #include "io.h"
 
-/* by LEGACY_FLAC we mean pre-1.1.3, before FLAC__SeekableStreamDecoder was merged into FLAC__StreamDecoder */
+/* By LEGACY_FLAC we mean pre-1.1.3, before FLAC__SeekableStreamDecoder
+ * was merged into FLAC__StreamDecoder. */
 #if !defined(FLAC_API_VERSION_CURRENT) || FLAC_API_VERSION_CURRENT < 8
 #define LEGACY_FLAC
 #else
@@ -49,14 +51,15 @@ struct flac_data
 #endif
 	struct io_stream *stream;
 	int bitrate;
+	int avg_bitrate;
 	int abort; /* abort playing (due to an error) */
-	
+
 	unsigned length;
-	unsigned total_samples;
-	
+	FLAC__uint64 total_samples;
+
 	FLAC__byte sample_buffer[SAMPLE_BUFFER_SIZE];
 	unsigned sample_buffer_fill;
-	
+
 	/* sound parameters */
 	unsigned bits_per_sample;
 	unsigned sample_rate;
@@ -113,8 +116,7 @@ static size_t pack_pcm_signed (FLAC__byte *data,
 		}
 	}
 
-	debug ("Converted %d bytes",
-			wide_samples * channels * bytes_per_sample);
+	debug ("Converted %u bytes", wide_samples * channels * bytes_per_sample);
 
 	return wide_samples * channels * bytes_per_sample;
 }
@@ -153,15 +155,13 @@ static void metadata_callback (
 
 	if (metadata->type == FLAC__METADATA_TYPE_STREAMINFO) {
 		debug ("Got metadata info");
-		
-		data->total_samples =
-			(unsigned)(metadata->data.stream_info.total_samples
-				   & 0xffffffff);
-		data->bits_per_sample =
-			metadata->data.stream_info.bits_per_sample;
+
+		data->total_samples = metadata->data.stream_info.total_samples;
+		data->bits_per_sample = metadata->data.stream_info.bits_per_sample;
 		data->channels = metadata->data.stream_info.channels;
 		data->sample_rate = metadata->data.stream_info.sample_rate;
-		data->length = data->total_samples / data->sample_rate;
+		if (data->total_samples > 0)
+			data->length = data->total_samples / data->sample_rate;
 	}
 }
 
@@ -174,7 +174,7 @@ static void error_callback (
 		FLAC__StreamDecoderErrorStatus status, void *client_data)
 {
 	struct flac_data *data = (struct flac_data *)client_data;
-	
+
 	if (status != FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC) {
 		debug ("Aborting due to error");
 		data->abort = 1;
@@ -195,7 +195,7 @@ static FLAC__StreamDecoderReadStatus read_callback (
 {
 	struct flac_data *data = (struct flac_data *)client_data;
 	ssize_t res;
-	
+
 	res = io_read (data->stream, buffer, *bytes);
 
 	if (res > 0) {
@@ -206,7 +206,7 @@ static FLAC__StreamDecoderReadStatus read_callback (
 		return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
 #endif
 	}
-	
+
 	if (res == 0) {
 		*bytes = 0;
 		/* not sure why this works, but if it ain't broke... */
@@ -233,7 +233,7 @@ static FLAC__StreamDecoderSeekStatus seek_callback (
 #endif
 {
 	struct flac_data *data = (struct flac_data *)client_data;
-	
+
 #ifdef LEGACY_FLAC
 	return io_seek(data->stream, absolute_byte_offset, SEEK_SET) >= 0
 		? FLAC__SEEKABLE_STREAM_DECODER_SEEK_STATUS_OK
@@ -275,9 +275,19 @@ static FLAC__StreamDecoderLengthStatus length_callback (
 		FLAC__uint64 *stream_length, void *client_data)
 #endif
 {
+	off_t file_size;
 	struct flac_data *data = (struct flac_data *)client_data;
 
-	*stream_length = io_file_size (data->stream);
+	file_size = io_file_size (data->stream);
+	if (file_size == -1)
+#ifdef LEGACY_FLAC
+		return FLAC__SEEKABLE_STREAM_DECODER_LENGTH_STATUS_ERROR;
+#else
+		return FLAC__STREAM_DECODER_LENGTH_STATUS_ERROR;
+#endif
+
+	*stream_length = file_size;
+
 #ifdef LEGACY_FLAC
 	return FLAC__SEEKABLE_STREAM_DECODER_LENGTH_STATUS_OK;
 #else
@@ -304,34 +314,32 @@ static void *flac_open_internal (const char *file, const int buffered)
 
 	data = (struct flac_data *)xmalloc (sizeof(struct flac_data));
 	decoder_error_init (&data->error);
-	
+
 	data->decoder = NULL;
 	data->bitrate = -1;
+	data->avg_bitrate = -1;
 	data->abort = 0;
 	data->sample_buffer_fill = 0;
 	data->last_decode_position = 0;
+	data->length = -1;
+	data->ok = 0;
 
 	data->stream = io_open (file, buffered);
 	if (!io_ok(data->stream)) {
 		decoder_error (&data->error, ERROR_FATAL, 0,
-				"Can't load file: %s",
-				io_strerror(data->stream));
-		io_close (data->stream);
+				"Can't load file: %s", io_strerror(data->stream));
 		return data;
 	}
 
-	data->ok = 1;
-
 #ifdef LEGACY_FLAC
 	if (!(data->decoder = FLAC__seekable_stream_decoder_new())) {
 		decoder_error (&data->error, ERROR_FATAL, 0,
 				"FLAC__seekable_stream_decoder_new() failed");
-		data->ok = 0;
 		return data;
 	}
 
 	FLAC__seekable_stream_decoder_set_md5_checking (data->decoder, false);
-	
+
 	FLAC__seekable_stream_decoder_set_metadata_ignore_all (data->decoder);
 	FLAC__seekable_stream_decoder_set_metadata_respond (data->decoder,
 			FLAC__METADATA_TYPE_STREAMINFO);
@@ -357,7 +365,6 @@ static void *flac_open_internal (const char *file, const int buffered)
 			!= FLAC__SEEKABLE_STREAM_DECODER_OK) {
 		decoder_error (&data->error, ERROR_FATAL, 0,
 				"FLAC__seekable_stream_decoder_init() failed");
-		data->ok = 0;
 		return data;
 	}
 
@@ -366,19 +373,17 @@ static void *flac_open_internal (const char *file, const int buffered)
 		decoder_error (&data->error, ERROR_FATAL, 0,
 				"FLAC__seekable_stream_decoder_process_until_end_of_metadata()"
 				" failed.");
-		data->ok = 0;
 		return data;
 	}
 #else
 	if (!(data->decoder = FLAC__stream_decoder_new())) {
 		decoder_error (&data->error, ERROR_FATAL, 0,
 				"FLAC__stream_decoder_new() failed");
-		data->ok = 0;
 		return data;
 	}
 
 	FLAC__stream_decoder_set_md5_checking (data->decoder, false);
-	
+
 	FLAC__stream_decoder_set_metadata_ignore_all (data->decoder);
 	FLAC__stream_decoder_set_metadata_respond (data->decoder,
 			FLAC__METADATA_TYPE_STREAMINFO);
@@ -387,19 +392,29 @@ static void *flac_open_internal (const char *file, const int buffered)
 			!= FLAC__STREAM_DECODER_INIT_STATUS_OK) {
 		decoder_error (&data->error, ERROR_FATAL, 0,
 				"FLAC__stream_decoder_init() failed");
-		data->ok = 0;
 		return data;
 	}
 
 	if (!FLAC__stream_decoder_process_until_end_of_metadata(data->decoder)) {
 		decoder_error (&data->error, ERROR_FATAL, 0,
-				"FLAC__stream_decoder_process_until_end_of_metadata()"
-				" failed.");
-		data->ok = 0;
+				"FLAC__stream_decoder_process_until_end_of_metadata() failed.");
 		return data;
 	}
 #endif
 
+	data->ok = 1;
+
+	if (data->length > 0) {
+		off_t data_size = io_file_size (data->stream);
+		if (data_size > 0) {
+			FLAC__uint64 pos;
+
+			if (FLAC__stream_decoder_get_decode_position (data->decoder, &pos))
+				data_size -= pos;
+			data->avg_bitrate = data_size * 8 / data->length;
+		}
+	}
+
 	return data;
 }
 
@@ -412,19 +427,18 @@ static void flac_close (void *void_data)
 {
 	struct flac_data *data = (struct flac_data *)void_data;
 
-	if (data->ok) {
-		if (data->decoder) {
+	if (data->decoder) {
 #ifdef LEGACY_FLAC
-			FLAC__seekable_stream_decoder_finish (data->decoder);
-			FLAC__seekable_stream_decoder_delete (data->decoder);
+		FLAC__seekable_stream_decoder_finish (data->decoder);
+		FLAC__seekable_stream_decoder_delete (data->decoder);
 #else
-			FLAC__stream_decoder_finish (data->decoder);
-			FLAC__stream_decoder_delete (data->decoder);
+		FLAC__stream_decoder_finish (data->decoder);
+		FLAC__stream_decoder_delete (data->decoder);
 #endif
-		}
-		io_close (data->stream);
 	}
 
+	io_close (data->stream);
+	decoder_error_clear (&data->error);
 	free (data);
 }
 
@@ -443,7 +457,7 @@ static void fill_tag (FLAC__StreamMetadata_VorbisComment_Entry *comm,
 	strncpy (name, (char *)comm->entry, eq - comm->entry);
 	name[eq - comm->entry] = 0;
 	value_length = comm->length - (eq - comm->entry + 1);
-	
+
 	if (value_length == 0) {
 		free (name);
 		return;
@@ -477,7 +491,7 @@ static void get_vorbiscomments (const char *filename, struct file_tags *tags)
 	FLAC__bool got_vorbis_comments = false;
 
 	debug ("Reading comments for %s", filename);
-	
+
 	if (!iterator) {
 		logit ("FLAC__metadata_simple_iterator_new() failed.");
 		return;
@@ -494,7 +508,7 @@ static void get_vorbiscomments (const char *filename, struct file_tags *tags)
 		if (FLAC__metadata_simple_iterator_get_block_type(iterator)
 				== FLAC__METADATA_TYPE_VORBIS_COMMENT) {
 			FLAC__StreamMetadata *block;
-			
+
 			block = FLAC__metadata_simple_iterator_get_block (
 					iterator);
 			if (block) {
@@ -511,7 +525,7 @@ static void get_vorbiscomments (const char *filename, struct file_tags *tags)
 		}
 	} while (!got_vorbis_comments
 			&& FLAC__metadata_simple_iterator_next(iterator));
-	
+
 	FLAC__metadata_simple_iterator_delete(iterator);
 }
 
@@ -520,11 +534,11 @@ static void flac_info (const char *file_name, struct file_tags *info,
 {
 	if (tags_sel & TAGS_TIME) {
 		struct flac_data *data;
-		
-		if ((data = flac_open_internal(file_name, 0))) {
+
+		data = flac_open_internal (file_name, 0);
+		if (data->ok)
 			info->time = data->length;
-			flac_close (data);
-		}
+		flac_close (data);
 	}
 
 	if (tags_sel & TAGS_COMMENTS)
@@ -536,12 +550,13 @@ static int flac_seek (void *void_data, int sec)
 	struct flac_data *data = (struct flac_data *)void_data;
 	FLAC__uint64 target_sample;
 
-	if (sec < 0 || (unsigned)sec > data->length)
+	if ((unsigned)sec > data->length)
 		return -1;
 
-	target_sample = (FLAC__uint64)((sec/(double)data->length) *
-			(double)data->total_samples);
-	
+	target_sample = (FLAC__uint64)(((double)sec / (double)data->length) *
+	                               (double)data->total_samples);
+
+
 #ifdef LEGACY_FLAC
 	if (FLAC__seekable_stream_decoder_seek_absolute(data->decoder,
 				target_sample))
@@ -585,10 +600,10 @@ static int flac_decode (void *void_data, char *buf, int buf_len,
 	sound_params->channels = data->channels;
 
 	decoder_error_clear (&data->error);
-	
+
 	if (!data->sample_buffer_fill) {
 		debug ("decoding...");
-		
+
 #ifdef LEGACY_FLAC
 		if (FLAC__seekable_stream_decoder_get_state(data->decoder) == FLAC__SEEKABLE_STREAM_DECODER_END_OF_STREAM)
 #else
@@ -621,7 +636,7 @@ static int flac_decode (void *void_data, char *buf, int buf_len,
 		if (decode_position > data->last_decode_position) {
 			int bytes_per_sec = bytes_per_sample * data->sample_rate
 				* data->channels;
-		 
+
 			data->bitrate = (decode_position
 				- data->last_decode_position) * 8.0
 				/ (data->sample_buffer_fill
@@ -652,21 +667,40 @@ static int flac_get_bitrate (void *void_data)
 	return data->bitrate;
 }
 
-static int flac_get_duration (void *void_data)
+static int flac_get_avg_bitrate (void *void_data)
 {
 	struct flac_data *data = (struct flac_data *)void_data;
 
-	return data->length;
+	return data->avg_bitrate / 1000;
+}
+
+static int flac_get_duration (void *void_data)
+{
+	int result = -1;
+	struct flac_data *data = (struct flac_data *)void_data;
+
+	if (data->ok)
+		result = data->length;
+
+	return result;
 }
 
 static void flac_get_name (const char *file ATTR_UNUSED, char buf[4])
 {
-	strcpy (buf, "FLA");
+	strcpy (buf, "FLC");
 }
 
 static int flac_our_format_ext (const char *ext)
 {
-	return !strcasecmp(ext, "flac") || !strcasecmp(ext, "fla");
+	return !strcasecmp (ext, "flac") || !strcasecmp (ext, "fla");
+}
+
+static int flac_our_format_mime (const char *mime)
+{
+	return !strcasecmp (mime, "audio/flac") ||
+	       !strncasecmp (mime, "audio/flac;", 11) ||
+	       !strcasecmp (mime, "audio/x-flac") ||
+	       !strncasecmp (mime, "audio/x-flac;", 13);
 }
 
 static void flac_get_error (void *prv_data, struct decoder_error *error)
@@ -691,10 +725,11 @@ static struct decoder flac_decoder = {
 	flac_get_duration,
 	flac_get_error,
 	flac_our_format_ext,
-	NULL,
+	flac_our_format_mime,
 	flac_get_name,
 	NULL,
-	NULL
+	NULL,
+	flac_get_avg_bitrate
 };
 
 struct decoder *plugin_init ()
